from typing import Dict

from dot_access import DotAccess
from expression import Expression
from index_access import IndexAccess
from literal import Literal
from no_interpretation_expression import NoInterpretationExpression
from reference import Reference
from sequence import Sequence


class Assignments:
    def __init__(self):
        self.__assignments: Dict[Reference, Expression] = dict()

    @property
    def assignments(self):
        return dict(self.__assignments)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Assignments):
            return False

        self_assignments = self.assignments
        other_assignments = other.assignments
        if len(self_assignments) != len(other_assignments):
            return False

        for key, value in self_assignments.items():
            if key in other_assignments:
                if not value == other_assignments[key]:
                    return False
            else:
                return False
        return True

    def __contains__(self, reference: Reference) -> bool:
        return reference in self.__assignments

    def __getitem__(self, reference: Reference) -> Expression:
        return self.__assignments[reference]

    def add(self, reference: Reference, expression: Expression) -> None:
        assert isinstance(reference, Reference)
        if self.__contains__(reference):
            raise ValueError

        self.__assignments[reference] = expression

    def __repr__(self) -> str:
        strings = []
        for key, value in self.__assignments.items():
            strings.append(str(key) + " -> " + str(value))
        return "{" + ", ".join(strings) + "}"

    def all_resolved(self) -> bool:
        for value in self.__assignments.values():
            if not value.is_closed():
                return False
        return True

    @staticmethod
    def substitute_in_no_interpretation_expression(
            no_interpretation_expression: NoInterpretationExpression
    ) -> NoInterpretationExpression:
        return no_interpretation_expression

    @staticmethod
    def substitute_in_literal(literal: Literal) -> Literal:
        return literal

    def substitute_in_reference(self, reference: Reference) -> Expression:
        if self.__contains__(reference):
            return_value = self.__getitem__(reference)
        else:
            return_value = reference
        return return_value

    def substitute_in_dot_access(self, dot_access: DotAccess) -> Expression:
        expression = self.substitute_in_expression(dot_access.expression)
        return DotAccess(expression, dot_access.name)

    def substitute_in_index_access(self, index_access: IndexAccess) -> Expression:
        expression = self.substitute_in_expression(index_access.expression)
        indices = list(map(self.substitute_in_expression, index_access.indices))
        return IndexAccess(expression, indices)

    def substitute_in_sequence(self, sequence: Sequence) -> Sequence:
        if len(self.__assignments) == 0:
            return sequence

        result = map(lambda x: self.substitute_in_expression(x),
                     sequence.sequence)
        # result = list()
        # for x in sequence.sequence:
        #     result.append(self.substitute_in_expression(x))

        sequence_type = type(sequence)
        return sequence_type(result)

    def substitute_in_expression(self, expression: Expression) -> Expression:
        if len(self.__assignments) == 0:
            return expression

        if isinstance(expression, NoInterpretationExpression):
            return Assignments.substitute_in_no_interpretation_expression(expression)
        if isinstance(expression, Literal):
            return Assignments.substitute_in_literal(expression)
        if isinstance(expression, Reference):
            return self.substitute_in_reference(expression)
        if isinstance(expression, DotAccess):
            return self.substitute_in_dot_access(expression)
        if isinstance(expression, IndexAccess):
            return self.substitute_in_index_access(expression)
        if isinstance(expression, Sequence):
            return self.substitute_in_sequence(expression)

        raise ValueError('Unknown type for Expression ' + str(type(expression)))

    def substitute_in_assignments(self, assignments):
        if len(self.__assignments) == 0:
            return assignments  # TODO: shallow copy iso returning same object?

        return_value: Assignments = Assignments()
        for key, value in self.__assignments.items():
            return_value.add(key, assignments.substitute_in_expression(value))
        return return_value
