from typing import List
from xml.etree import ElementTree
from xml.etree.ElementTree import Element

from argument import Argument
from arguments import Arguments
from concatenation import Concatenation
from dot_access import DotAccess
from expression import Expression
from file import File
from file_path import FilePath
from file_range import parse
from files import Files
from index_access import IndexAccess
from init_inspection_object_common import InitInspectionObjectCommon
from literal import Literal
from no_interpretation_expression import NoInterpretationExpression
from name import Name
from param_assignment import ParamAssignment
from parameter import Parameter
from parameter_declaration import ParameterDeclaration
from parameter_declarations import ParameterDeclarations
from subprogram_declaration import SubprogramDeclaration
from variable import Variable
from xml_extraction import \
    get_single_child_node, \
    get_single_child_node_with_name, \
    get_text_of_node


def _extract_variable(variable_node: Element) -> Variable:
    return Variable(Name(variable_node.attrib['name']),
                    FilePath(variable_node.attrib['file']),
                    parse(variable_node.attrib['location']))


def _extract_parameter(parameter_node: Element) -> Parameter:
    return Parameter(Name(parameter_node.attrib['name']),
                     FilePath(parameter_node.attrib['file']),
                     parse(parameter_node.attrib['location']))


def _extract_parameter_declaration(parameter_node: Element) -> ParameterDeclaration:
    return ParameterDeclaration(Name(parameter_node.attrib['name']),
                                FilePath(parameter_node.attrib['file']),
                                parse(parameter_node.attrib['location']),
                                Name(get_text_of_node(parameter_node)))


def _extract_parameter_declarations(file_path: FilePath, subprogram_node: Element) \
        -> ParameterDeclarations:
    parameters = ParameterDeclarations()
    for parameter_node in subprogram_node.findall('parameter'):
        parameter = _extract_parameter_declaration(parameter_node)
        assert file_path == parameter.file_path
        parameters.add(parameter)
    return parameters


def _extract_subprogram_declarations(file_node: Element) -> List[SubprogramDeclaration]:
    return_value = []
    file_path = FilePath(file_node.attrib['name'])
    for subprogram_node in file_node.findall('subprogram'):
        assert file_path == FilePath(subprogram_node.attrib['file'])
        parameters = _extract_parameter_declarations(file_path, subprogram_node)
        subprogram_declaration = SubprogramDeclaration(Name(subprogram_node.attrib['name']),
                                                       file_path,
                                                       parse(subprogram_node.attrib['location']),
                                                       parameters)
        return_value.append(subprogram_declaration)
    return return_value


def _extract_concat(concat_node: Element) -> Concatenation:
    concatenation = Concatenation()
    for child_node in concat_node:
        expression = _extract_expression(child_node)
        concatenation.append(expression)
    return concatenation


def _extract_dot(dot_node: Element) -> DotAccess:
    assert len(dot_node) == 2
    assert dot_node[1].tag == 'component'
    expression = _extract_expression(dot_node[0])
    name = Name(dot_node[1].attrib['name'])
    return DotAccess(expression, name)


def _extract_call(call_node: Element) -> IndexAccess:
    length = len(call_node)
    assert length >= 2, "unexpected length = " + str(length)
    expression = _extract_expression(call_node[0])
    indices = list()
    for index in range(1, length):
        indices.append(_extract_expression(call_node[index]))
    return IndexAccess(expression, indices)


def _extract_expression(expression_node: Element) -> Expression:
    if expression_node.tag == 'literal':
        return Literal(get_text_of_node(expression_node))

    if expression_node.tag == 'parameter':
        return _extract_parameter(expression_node)

    if expression_node.tag == 'variable':
        return _extract_variable(expression_node)

    if expression_node.tag == 'concat':
        return _extract_concat(expression_node)

    if expression_node.tag == 'dot':
        return _extract_dot(expression_node)

    if expression_node.tag == 'call':
        return _extract_call(expression_node)

    if expression_node.tag == 'component':
        return _extract_variable(expression_node)   # todo Make special component basis expression?

    raise ValueError('unknown node tag "' + expression_node.tag + '"')

    
def _extract_argument(argument_node: Element) -> Argument:
    name = Name(argument_node.attrib['name'])
    assert len(argument_node) == 1
    argument_expression = _extract_expression(argument_node[0])
    return Argument(name, argument_expression)


def _extract_arguments(call_node: Element) -> Arguments:
    arguments = Arguments()
    for argument_node in call_node.findall('argument'):
        argument = _extract_argument(argument_node)
        arguments.add(argument)
    return arguments


def _extract_init_inspection_object_commons(subprogram_node: Element) \
        -> List[InitInspectionObjectCommon]:
    return_value = []
    for init_inspection_object_common_node in \
            subprogram_node.findall('Init_Inspection_Object_Common'):
        class_expression_node = get_single_child_node_with_name(init_inspection_object_common_node,
                                                                'class')
        class_expression = _extract_expression(get_single_child_node(class_expression_node))
        name_node = get_single_child_node_with_name(init_inspection_object_common_node,
                                                    'name')
        name_expression = _extract_expression(get_single_child_node(name_node))
        init_inspection_object_common = InitInspectionObjectCommon(
            class_expression,
            name_expression
        )
        return_value.append(init_inspection_object_common)
    return return_value


def _extract_expression_from_child(node: Element) -> Expression:
    if len(node) == 0:
        return NoInterpretationExpression()
    return _extract_expression(get_single_child_node(node))


def _extract_param_assignments(subprogram_node: Element) \
        -> List[ParamAssignment]:
    return_value = []
    for param_assignment_node in \
            subprogram_node.findall('Param_Assignment'):
        app_node = get_single_child_node_with_name(param_assignment_node, 'app')
        app_expression = _extract_expression_from_child(app_node)
        attribute_node = get_single_child_node_with_name(param_assignment_node, 'attribute')
        attribute_expression = _extract_expression_from_child(attribute_node)
        value_node = get_single_child_node_with_name(param_assignment_node, 'value')
        value_expression = _extract_expression_from_child(value_node)
        param_assignment = ParamAssignment(
            app_expression,
            attribute_expression,
            value_expression,
            FilePath(param_assignment_node.attrib['file']),
            parse(param_assignment_node.attrib['location'])
        )
        return_value.append(param_assignment)
    return return_value


class ExtractSymbolicResult:
    def __extract_declarations(self, files_node: Element) -> None:
        for file_node in files_node.findall('file'):
            file = File(FilePath(file_node.attrib['name']))
            self.__files.add(file)
            for subprogram_declaration in _extract_subprogram_declarations(file_node):
                file.add_subprogram_declaration(subprogram_declaration)

    def __extract_specifics(self, file_node: Element) -> None:
        file = self.__files[FilePath(file_node.attrib['name'])]
        for subprogram_node in file_node.findall('subprogram'):
            subprogram_declaration: SubprogramDeclaration = \
                file[parse(subprogram_node.attrib['location'])]

            for init_inspection_object_common in \
                    _extract_init_inspection_object_commons(subprogram_node):
                subprogram_declaration. \
                    add_init_inspection_object_common(init_inspection_object_common)

    def __extract_calls(self, file_node: Element) -> None:
        caller_file = self.__files[FilePath(file_node.attrib['name'])]
        for caller_node in file_node.findall('subprogram'):
            caller: SubprogramDeclaration = \
                caller_file[parse(caller_node.attrib['location'])]
            for call_node in caller_node.findall('call'):
                self.__nr_of_subprogram_calls += 1
                callee_node = get_single_child_node_with_name(call_node, 'subprogram')
                file_path = FilePath(callee_node.attrib['file'])
                if file_path in self.__files:
                    callee_file = self.__files[file_path]
                    file_range = parse(callee_node.attrib['location'])
                    if file_range in callee_file:
                        callee = callee_file[file_range]
                        arguments = _extract_arguments(call_node)
                        # add missing list_of_arguments
                        for parameter in callee.parameter_declarations.parameter_declaration_list:
                            if parameter.name not in arguments:
                                arguments.add(Argument(parameter.name,
                                                       NoInterpretationExpression()), )
                        caller.add_call(callee, arguments, parse(call_node.attrib['location']))
                    # else:
                    #     print("MISSING SUBPROGRAM " + str(file_path) + " - " + str(file_range))
                    # Note: abstract subprograms have no implementation!
                else:
                    file_path_string = str(file_path)
                    if not file_path_string.startswith("c:\\gnatpro\\21.0\\"):
                        print("MISSING FILE " + file_path_string)

    def __extract_references(self, root: Element) -> None:
        for file_node in root.findall('file'):
            self.__extract_specifics(file_node)
            self.__extract_calls(file_node)

    def __extract_param_assignments(self, root: Element) -> None:
        for file_node in root.findall('file'):
            file = self.__files[FilePath(file_node.attrib['name'])]
            for subprogram_node in file_node.findall('subprogram'):
                subprogram_declaration: SubprogramDeclaration = \
                    file[parse(subprogram_node.attrib['location'])]

                for param_assignment in \
                        _extract_param_assignments(subprogram_node):
                    subprogram_declaration. \
                        add_param_assignment(param_assignment)

    # first we extract all declarations (files, subprograms)
    # second we extract all relations
    # (while checking consistency: declared before used)
    def __init__(self):
        tree = ElementTree. \
            parse("C:/bright/git/ada/Inspection_Analysis/src.new/Result_Symbolic.Xml")
        root = tree.getroot()
        self.__files = Files()
        self.__nr_of_subprogram_calls = 0
        self.__extract_declarations(root)
        self.__extract_references(root)
        self.__extract_param_assignments(root)

    @property
    def files(self) -> Files:
        return self.__files

    @property
    def nr_of_subprogram_calls(self) -> int:
        return self.__nr_of_subprogram_calls


def main():
    extract_symbolic_result = ExtractSymbolicResult()
    files: Files = extract_symbolic_result.files
    print("Nr Of Files = " + str(len(files)))
    print("Nr Of Subprograms = " + str(files.nr_of_subprogram_declarations()))
    print("Nr Of Calls = " + str(extract_symbolic_result.nr_of_subprogram_calls))
    print("Nr Of Remaining Calls = " + str(files.nr_of_subprogram_calls()))
    print("Nr Of Init Inspection Object Commons = " +
          str(files.nr_of_init_inspection_object_commons()))


if __name__ == "__main__":
    main()
