from typing import List, Dict, Set
from parameter_declaration import ParameterDeclaration
from name import Name


class ParameterDeclarations:
    """
    The parameter declarations of a single function.
    The name of a parameter declaration is unique within the scope of this single function.
    The addition of a parameter declaration with a name that is already added earlier will fail.
    The position of a parameter declaration is fixed.
    """
    def __init__(self):
        self.__parameter_declaration_list: List[ParameterDeclaration] = list()
        self.__parameter_declaration_dict: Dict[Name, ParameterDeclaration] = dict()

    def __contains__(self, name: Name) -> bool:
        return name in self.__parameter_declaration_dict

    @property
    def parameter_declaration_list(self) -> List[ParameterDeclaration]:
        return list(self.__parameter_declaration_list)

    def add(self, parameter_declaration: ParameterDeclaration) -> None:
        """
        Add a parameter_declaration declaration to the parameter_declaration declarations.
        The addition of a parameter_declaration declaration with a name that is already
        added earlier will fail.
        """
        if self.__contains__(parameter_declaration.name):
            raise ValueError('Name of Parameter Declaration "'
                             + str(parameter_declaration.name)
                             + '" already in parameter declarations')

        self.__parameter_declaration_list.append(parameter_declaration)
        self.__parameter_declaration_dict[parameter_declaration.name] = parameter_declaration

    def get_by_name(self, name: Name) -> ParameterDeclaration:
        return self.__parameter_declaration_dict[name]

    def get_by_position(self, index: int) -> ParameterDeclaration:
        assert index >= 0
        assert index < len(self.__parameter_declaration_list)
        return self.__parameter_declaration_list[index]

    def type_names(self) -> Set[Name]:
        return_value = set()
        for parameter_declaration in self.__parameter_declaration_list:
            return_value.add(parameter_declaration.type_name)
        return return_value

    def __len__(self) -> int:
        return len(self.__parameter_declaration_dict)

    def __repr__(self) -> str:
        return "(" + ", ".join(map(str, self.__parameter_declaration_list)) + ")"
