from arguments import Arguments
from dataclasses import dataclass
from file_path import FilePath
from file_range import FileRange
from init_inspection_object_common import InitInspectionObjectCommon
from name import Name
from param_assignment import ParamAssignment
from parameter_declarations import ParameterDeclarations
from xml_generation import xml_node


class SubprogramDeclaration:
    @dataclass(frozen=True)
    class SubprogramCall:
        caller: object  # Unfortunately, can not specify SubprogramDeclaration as type
        callee: object  # Unfortunately, can not specify SubprogramDeclaration as type
        arguments: Arguments
        file_range: FileRange

        def __post_init__(self):
            if not isinstance(self.caller, SubprogramDeclaration):
                raise ValueError('Type of caller must be SubprogramDeclaration, but is '
                                 + str(type(self.caller)))
            if not isinstance(self.callee, SubprogramDeclaration):
                raise ValueError('Type of callee must be SubprogramDeclaration, but is '
                                 + str(type(self.callee)))

        def to_xml(self) -> str:
            attributes: dict[str, str] = {"callee": str(self.callee.name),
                                          "caller": str(self.caller.name),
                                          "file": str(self.caller.file_path),
                                          "location": str(self.file_range)}
            return xml_node("call", attributes=attributes)

    def __init__(self,
                 name: Name,
                 file_path: FilePath,
                 file_range: FileRange,
                 parameter_declarations: ParameterDeclarations
                 ) -> None:
        self.__name: Name = name
        self.__file_path: FilePath = file_path
        self.__file_range: FileRange = file_range
        self.__parameter_declarations: ParameterDeclarations = parameter_declarations
        # calls to this subprogram declaration
        self.__to_calls: list[SubprogramDeclaration.SubprogramCall] = []
        # calls from/by this subprogram declarations
        self.__from_calls: list[SubprogramDeclaration.SubprogramCall] = []
        self.__init_inspection_object_commons: list[InitInspectionObjectCommon] = []
        self.__param_assignments: list[ParamAssignment] = []

    @property
    def name(self) -> Name:
        return self.__name

    @property
    def file_path(self) -> FilePath:
        return self.__file_path

    @property
    def file_range(self) -> FileRange:
        return self.__file_range

    @property
    def parameter_declarations(self) -> ParameterDeclarations:
        return self.__parameter_declarations

    @property
    def to_calls(self) -> list[SubprogramCall]:
        return list(self.__to_calls)

    @property
    def from_calls(self) -> list[SubprogramCall]:
        return list(self.__from_calls)

    @property
    def init_inspection_object_commons(self) -> list[InitInspectionObjectCommon]:
        return list(self.__init_inspection_object_commons)

    @property
    def param_assignments(self) -> list[ParamAssignment]:
        return list(self.__param_assignments)

    def __repr__(self) -> str:
        return str(self.__name) + str(self.__parameter_declarations) + \
               "[" + str(self.__file_path) + "@" + str(self.__file_range) + "]"

    def add_call(self,
                 other: object,  # Unfortunately, can not specify SubprogramDeclaration as type
                 arguments: Arguments,
                 file_range: FileRange
                 ) -> None:
        if not isinstance(other, SubprogramDeclaration):
            raise ValueError('Type of other must be SubprogramDeclaration, but is '
                             + str(type(other)))

        if not arguments.match(other.parameter_declarations):
            raise ValueError("Arguments don't match parameters of " +
                             "called SubprogramDeclaration (other)\n" +
                             str(arguments) + "\nversus\n" +
                             str(other.parameter_declarations))

        subprogram_call = SubprogramDeclaration.SubprogramCall(self, other, arguments, file_range)
        self.__from_calls.append(subprogram_call)
        other.__to_calls.append(subprogram_call)

    def add_init_inspection_object_common(self,
                                          init_inspection_object_common: InitInspectionObjectCommon
                                          ) -> None:
        self.__init_inspection_object_commons.append(init_inspection_object_common)

    def add_param_assignment(self,
                             param_assignment: ParamAssignment
                             ) -> None:
        self.__param_assignments.append(param_assignment)
